'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _Patterns = require('./Patterns');

var _Patterns2 = _interopRequireDefault(_Patterns);

var _StorageMap = require('./StorageMap');

var _StorageMap2 = _interopRequireDefault(_StorageMap);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Ely
 *
 * Maps and unmaps object properties to emitter.
 * Every further change of properties will emit the changes.
 *
 * Can keep the emitter inside or outside of the original object.
 */
var Ely = function () {
  function Ely() {
    _classCallCheck(this, Ely);
  }

  _createClass(Ely, null, [{
    key: 'mapIn',

    /**
     * Maps and keeps emitter inside the object
     *
     * Shortkey for map(object, true, patterns)
     *
     * @param {Object} object object with properties
     * @param {Patterns} patterns patterns of matching
     * @return {EventEmitter} event emitter
     */
    value: function mapIn(object) {
      var patterns = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;

      return Ely.map(object, true, patterns);
    }

    /**
     * Maps and keeps emitter outside the object
     *
     * Shortkey for map(object, false, patterns)
     *
     * @param {Object} object object with properties
     * @param {Patterns} patterns patterns of matching
     * @return {EventEmitter} event emitter
     */

  }, {
    key: 'mapOut',
    value: function mapOut(object) {
      var patterns = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;

      return Ely.map(object, false, patterns);
    }

    /**
     * Maps object properties
     *
     * Main map method
     *
     * @param {Object} object object with properties
     * @param {boolean} embed embed the emitter or not
     * @param {Patterns} patterns patterns of matching
     * @return {EventEmitter} event emitter
     */

  }, {
    key: 'map',
    value: function map(object) {
      var embed = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
      var patterns = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;

      var defaultPatterns = Ely.defaultPatterns;
      patterns = patterns === undefined ? defaultPatterns : patterns;

      var emitter = void 0;

      if (embed) {
        if (object.ely !== undefined) {
          if (object.ely instanceof _events2.default) {
            emitter = object.ely;
          } else {
            embed = false;
          }
        }
      }

      if (emitter === undefined) {
        emitter = new _events2.default();
      }

      if (embed && object.ely === undefined) {
        Object.defineProperty(object, 'ely', {
          configurable: true,
          enumerable: true,
          writable: false,
          value: emitter
        });
      }

      var elyMapKey = Ely.findElyMapKey(object);
      var elyMap = object[elyMapKey] !== undefined ? object[elyMapKey] : new _StorageMap2.default();
      var elyMapKeys = Object.keys(elyMap.map);
      elyMapKeys.push(elyMapKey);

      var emitall = patterns !== undefined ? patterns.emitall : false;

      Object.keys(object).forEach(function (key) {
        if (elyMapKeys.indexOf(key) === -1) {
          var descriptor = Object.getOwnPropertyDescriptor(object, key);
          if (descriptor.writable) {
            var valueType = _typeof(descriptor.value);
            if (valueType !== 'function') {
              if (patterns === undefined || Ely.match(key, patterns.accept) && Ely.notmatch(key, patterns.skip)) {
                var privateKey = Ely.findEmptyKey(elyMap.map, key, '_');
                var value = object[key];

                Object.defineProperty(elyMap.values, privateKey, {
                  configurable: true,
                  enumerable: true,
                  writable: true,
                  value: value
                });
                Object.defineProperty(object, key, {
                  configurable: true,
                  enumerable: true,
                  set: function set(value) {
                    emitter.emit(key, value, elyMap.values[privateKey]);
                    if (emitall) {
                      emitter.emit('*', value, elyMap.values[privateKey], key);
                    }
                    elyMap.values[privateKey] = value;
                  },
                  get: function get() {
                    return elyMap.values[privateKey];
                  }
                });

                elyMap.map[key] = privateKey;
              }
            }
          }
        }
      });

      Object.defineProperty(object, elyMapKey, {
        configurable: true,
        enumerable: false,
        writable: true,
        value: elyMap
      });

      return emitter;
    }

    /**
     * Unmaps object properties
     *
     * @param {Object} object object with properties
     */

  }, {
    key: 'unmap',
    value: function unmap(object) {
      var elyMapKey = Ely.findElyMapKey(object);
      var elyMap = object[elyMapKey];

      if (elyMap !== undefined) {
        Object.keys(elyMap.map).forEach(function (key) {
          Object.defineProperty(object, key, {
            configurable: true,
            enumerable: true,
            writable: true,
            value: elyMap.values[elyMap.map[key]]
          });
        });

        delete object[elyMapKey];
        if (object['ely'] !== undefined && object['ely'] instanceof _events2.default) {
          delete object['ely'];
        }
      }
    }

    /**
     * Gets default pattern
     *
     * @return {Patterns} default patterns
     */

  }, {
    key: 'findElyMapKey',


    /**
     * Finds the ely map key
     *
     * Checks if default ely map key already exists and replaces
     * it with unique one.
     *
     * @param {Object} object object with properties
     * @return {string}
     * @protected
     */
    value: function findElyMapKey(object) {
      var name = '__ely_map__';
      while (object[name] !== undefined && !(object[name] instanceof _StorageMap2.default)) {
        name = '_' + name + '_';
      }
      if (object[name] === undefined) {
        return name;
      } else if (object[name] instanceof _StorageMap2.default) {
        return name;
      } else {
        return undefined;
      }
    }

    /**
     * Finds empty key
     *
     * Checks if key overlap inside object. Creates unique one if needed.
     *
     * @param {Object} object object with properties
     * @param {string} key key to be checked
     * @param {string} prefix prefix to be applied
     * @return {string} unique key
     * @protected
     */

  }, {
    key: 'findEmptyKey',
    value: function findEmptyKey(object, key, prefix) {
      var name = key;
      while (object[name] !== undefined) {
        name = prefix + name;
      }
      return name;
    }

    /**
     * Matches key to pattern
     *
     * Checks if key matches a given pattern
     *
     * @param {string} key key to be checked
     * @param {Patterns} patterns patterns to check against
     * @return {boolean} match result
     * @protected
     */

  }, {
    key: 'match',
    value: function match(key, patterns) {
      if (patterns === undefined) {
        return true;
      } else if (Array.isArray(patterns)) {
        var length = patterns.length;
        var i = void 0;
        var pass = false;
        for (i = 0; i < length; ++i) {
          var pattern = patterns[i];
          if ((typeof pattern === 'undefined' ? 'undefined' : _typeof(pattern)) === 'object' && pattern.constructor !== undefined && pattern.constructor.name === 'RegExp') {
            pass = pass || pattern.test(key);
          }
        }
        return pass;
      } else if ((typeof patterns === 'undefined' ? 'undefined' : _typeof(patterns)) === 'object' && patterns.constructor !== undefined && patterns.constructor.name === 'RegExp') {
        return patterns.test(key);
      } else {
        return true;
      }
    }

    /**
     * No Matches key to pattern
     *
     * Checks if key does not match a given pattern
     *
     * @param {string} key key to be checked
     * @param {Patterns} patterns patterns to check against
     * @return {boolean} match result
     * @protected
     */

  }, {
    key: 'notmatch',
    value: function notmatch(key, patterns) {
      if (patterns === undefined) {
        return true;
      } else if (Array.isArray(patterns)) {
        var length = patterns.length;
        var i = void 0;
        var pass = false;
        for (i = 0; i < length; ++i) {
          var pattern = patterns[i];
          if ((typeof pattern === 'undefined' ? 'undefined' : _typeof(pattern)) === 'object' && pattern.constructor !== undefined && pattern.constructor.name === 'RegExp') {
            pass = pass || pattern.test(key);
          }
        }
        return !pass;
      } else if ((typeof patterns === 'undefined' ? 'undefined' : _typeof(patterns)) === 'object' && patterns.constructor !== undefined && patterns.constructor.name === 'RegExp') {
        return !patterns.test(key);
      } else {
        return true;
      }
    }
  }, {
    key: 'defaultPatterns',
    get: function get() {
      return new _Patterns2.default(undefined, /^_.*$/, true);
    }
  }]);

  return Ely;
}();

exports.default = Ely;